# SPDX-License-Identifier: GPL-3.0-only

from multiprocessing import Process, Queue

import numpy as np
import pyaudio
import pyrubberband


class AudioPlayer:

    CHUNK_SIZE = 1024

    def __init__(self):
        self.cmd_queue = Queue()
        self.data_queue = Queue()
        self.status_queue = Queue()
        self.is_playing = False
        self._playback_time = 0
        self.effects_process = None
        self.play_process = None

    @property
    def playback_time(self):
        self.drain_status_queue()
        return self._playback_time

    def drain_status_queue(self):
        while not self.status_queue.empty():
            status_msg = self.status_queue.get()
            if "time" in status_msg:
                self._playback_time = status_msg["time"]

    def _recreate_queues(self):
        self.cmd_queue.close()
        self.data_queue.close()
        self.status_queue.close()
        self.cmd_queue = Queue()
        self.data_queue = Queue()
        self.status_queue = Queue()

    def play(self, signal, sampling_rate, playback_speed=1.0):
        if self.is_playing:
            # TODO raise expection
            return
        self._recreate_queues()
        self.is_playing = True
        self.effects_process = Process(target=effects_process_wrapper,
                                       args=(signal, self.data_queue),
                                       kwargs={
                                           "sampling_rate": sampling_rate,
                                           "playback_speed": playback_speed})

        if len(signal.shape) == 2 and signal.shape[1] == 2:
            nr_channels = 2
        else:
            nr_channels = 1
        self.play_process = Process(target=playback_process_wrapper,
                                    args=(self.cmd_queue, self.data_queue,
                                          self.status_queue,
                                          sampling_rate, nr_channels))
        print("Starting effects process")
        self.effects_process.start()
        self.play_process.start()

    def stop(self):
        self.cmd_queue.put("STOP")
        if self.effects_process is not None:
            self.effects_process.join()
            self.effects_process = None
        if self.play_process is not None:
            self.play_process.join()
            self.play_process = None
        self.is_playing = False


def effects_process_wrapper(signal, data_queue, sampling_rate=44100,
                            playback_speed=1.0):
    # If signal is large, process a smaller chunk first
    if len(signal)//sampling_rate > 20:  # large means longer than 20s
        chunk = pyrubberband.time_stretch(
            signal[0:sampling_rate*20],
            rate=playback_speed, sr=sampling_rate).astype(np.float32)/2
        data_queue.put(chunk)
    whole = pyrubberband.time_stretch(signal, rate=playback_speed,
                                      sr=sampling_rate).astype(np.float32)/2
    data_queue.put(whole)


def playback_process_wrapper(cmd_queue, data_queue, status_queue,
                             sampling_rate, nr_channels):
    print("playback started")
    time = 0
    pyaudio_obj = pyaudio.PyAudio()
    stream = pyaudio_obj.open(format=pyaudio.paFloat32,
                              channels=nr_channels,
                              rate=sampling_rate,
                              # TODO remove magic device number 6
                              # output_device_index=6,
                              output=True
                              )
    signal = []
    while True:
        while not data_queue.empty():
            signal = data_queue.get()
        if not cmd_queue.empty():
            cmd = cmd_queue.get()
            if cmd == "STOP":
                break
        if len(signal) > 0:
            chunk = signal[time:time+AudioPlayer.CHUNK_SIZE]
            stream.write(chunk.tobytes())
            time += AudioPlayer.CHUNK_SIZE
            status_queue.put({"time": time/sampling_rate})
            if time > len(signal):
                time = 0
    stream.close()



# SPDX-License-Identifier: GPL-3.0-only

import os

import soundfile as sf

from .audio import AudioPlayer
from .wave_display import AnimatedWaveDisplay

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # noqa


class App:

    def __init__(self):
        self.builder = Gtk.Builder()
        print(__file__)
        gladefile_path = os.path.join(os.path.dirname(__file__), "static",
                                      "main.glade")
        self.builder.add_from_file(gladefile_path)
        self.builder.connect_signals(self)
        main_window = self.builder.get_object("main_window")
        main_box = self.builder.get_object("main_box")
        self.wave_display = AnimatedWaveDisplay()
        self.wave_display.connect("selection-updated",
                                  self.on_selection_updated)
        main_box.pack_start(self.wave_display, True, True, 0)
        main_window.show_all()

        self.file_dialog = self.builder.get_object("open_dialog")
        self.file_dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                     Gtk.STOCK_OPEN, Gtk.ResponseType.OK)

        self.audio_player = AudioPlayer()
        self.marked_time = None

    def on_quit(self, *args):
        self.audio_player.stop()
        Gtk.main_quit()

    def on_selection_updated(self, wave_display, selection0, selection1):
        self.time_selection = (selection0, selection1)
        self.audio_player.stop()
        return True

    def on_open_clicked(self, *args):
        response = self.file_dialog.run()
        if response == Gtk.ResponseType.OK:
            path = self.file_dialog.get_filename()
            self.load_wav(path)
        self.file_dialog.hide()

    def on_main_window_key_press_event(self, window, event):
        def scroll_seconds(delta):
            self.wave_display.set_display_range(
                [i+delta for i in
                 self.wave_display.display_range])

        def zoom(zoom_factor):
            zoom_diff = (self.wave_display.display_range[1] -
                         self.wave_display.display_range[0])*zoom_factor

            self.wave_display.set_display_range(
                [self.wave_display.display_range[0] + zoom_diff,
                 self.wave_display.display_range[1] - zoom_diff]
            )

        if event.keyval == 65363:  # Right arrow
            scroll_seconds(1)
        elif event.keyval == 65361:  # Left arrow
            scroll_seconds(-1)
        elif event.keyval == 65362:  # Up arrow
            zoom(0.1)
        elif event.keyval == 65364:  # Down arrow
            zoom(-0.1)
        elif event.keyval == 97:  # a
            offset = (0 if self.time_selection is None
                      else self.time_selection[0])
            playback_time = offset + self.audio_player.playback_time
            if self.marked_time is None:
                self.marked_time = playback_time
            else:
                self.time_selection = sorted([self.marked_time, playback_time])
                self.wave_display.set_selection(self.time_selection)
                self.marked_time = None
                self.audio_player.stop()
        elif event.keyval == 32:  # Spacebar
            self.on_play_pause_clicked()
        else:
            return False
        return True

    def on_play_pause_clicked(self, *args):
        if self.audio_player.is_playing:
            self.audio_player.stop()
        else:
            speed = self.builder.get_object("speed_scale").get_value() / 100
            if self.time_selection:
                start_sample = int(self.time_selection[0]*self.sampling_rate)
                end_sample = int(self.time_selection[1]*self.sampling_rate)
            else:
                start_sample = 0
                end_sample = len(self.data)
            self.audio_player.play(self.data[start_sample:end_sample],
                                   self.sampling_rate, playback_speed=speed)

    def load_wav(self, path):
        self.data, self.sampling_rate = sf.read(path)
        self.wave_display.update_data(self.data, self.sampling_rate)
        # TODO remove magic numbers [0-10]
        self.time_selection = None
        self.wave_display.set_display_range([0, 10])
        self.wave_display.set_selection(None)
        self.audio_player.stop()

    def run(self):
        Gtk.main()



# SPDX-License-Identifier: GPL-3.0-only

"""Widget for interacting with waveforms."""
import bisect

from matplotlib.backends.backend_gtk3agg import (
    FigureCanvasGTK3Agg as FigureCanvas)
from matplotlib.figure import Figure

from gi.repository import Gtk, GObject


class AnimatedWaveDisplay(Gtk.Box):
    __gsignals__ = {
        'selection-updated': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,
                              (GObject.TYPE_FLOAT, GObject.TYPE_FLOAT))
    }

    def __init__(self):
        super().__init__()
        self.data = []
        self.timestamps = []
        self.display_range = [0, 10]
        self.selection = None

        self.fig = Figure(figsize=(5, 4), dpi=50)
        self.ax = self.fig.add_subplot(111)
        self.canvas = FigureCanvas(self.fig)

        self.pack_start(self.canvas, True, True, 0)
        self.canvas.mpl_connect('button_press_event',
                                self.on_mouse_press)
        self.canvas.mpl_connect('motion_notify_event',
                                self.on_mouse_movement)
        self.canvas.mpl_connect('button_release_event',
                                self.on_mouse_release)
        self.canvas.mpl_connect('draw_event', self.on_draw)

    def update_data(self, data, sampling_rate):
        # Convert data to mono if streo, no reason to display stereo
        if not(len(data.shape) == 1 or data.shape[1] == 1):
            data = data.T[0]
        self.data = data / data.max()
        # Timestamps are required to work with time data instead of sample nos.
        self.timestamps = [i/sampling_rate for i in range(len(data))]
        self.update_figure(redraw_data=True)

    def on_draw(self, *args):
        self.background = self.fig.canvas.copy_from_bbox(
            self.ax.get_figure().bbox)
        self.plot, = self.ax.plot([], [], animated=True)
        self.data_should_redraw = True
        self.update_figure(redraw_data=True)
        return False

    def update_figure(self, redraw_data=False):
        if redraw_data:
            self.fig.canvas.restore_region(self.background)
            start_index = bisect.bisect(self.timestamps, self.display_range[0])
            end_index = bisect.bisect(self.timestamps, self.display_range[1])
            # Show no more than 10_000 samples, for speedup
            step = max(1, (end_index-start_index) // 10_000)
            self.plot.set_data(self.timestamps[start_index:end_index:step],
                               self.data[start_index:end_index:step])
            self.ax.set_xlim(self.display_range)
            self.ax.set_ylim([-1, 1])
            self.ax.draw_artist(self.plot)
            self.plot_cache = self.fig.canvas.copy_from_bbox(
                self.ax.get_figure().bbox)

        if not redraw_data:
            self.fig.canvas.restore_region(self.plot_cache)
        if self.selection and self.selection[0] != self.selection[1]:
            selection_axvspan = self.ax.axvspan(*self.selection,
                                                color='red',
                                                alpha=0.5,
                                                animated=True)
            self.ax.draw_artist(selection_axvspan)
        self.fig.canvas.blit(self.fig.bbox)

    def set_display_range(self, display_range):
        self.display_range = display_range
        self.update_figure(redraw_data=True)

    def set_selection(self, selection):
        self.selection = selection
        self.update_figure()

    def on_mouse_press(self, event):
        self.set_selection([event.xdata, event.xdata])

    def on_mouse_movement(self, event):
        if event.button and event.button == event.button.LEFT:
            self.set_selection([self.selection[0], event.xdata])

    def on_mouse_release(self, event):
        self.selection.sort()
        self.emit("selection-updated", *self.selection)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mancho",
    version="0.0.1",
    author="Ali Rasim Kocal",
    author_email="arkocal@posteo.net",
    description="A musical transcription tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.arkocal.rocks/arkocal/transcriber",
    packages=["mancho"],
    package_data={"mancho": ["static/main.glade"]},
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    keywords="music musical transcription tool",
    python_requires='>=3.6',
    install_requires=["soundfile", "numpy", "pyrubberband"]
)
